"""smartcard.util package

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import annotations

import warnings

PACK = 1
HEX = 2
UPPERCASE = 4
COMMA = 8


def padd(bytelist: list[int], length: int, padding: str = "FF"):
    """Padds a byte list with a constant byte value (default is x0FF)
    @param bytelist: the byte list to padd
    @param length: the total length of the resulting byte list;
                  no padding if length is smaller than the byte list length
    @param padding: padding value (default is 0xff)

    @return: the padded bytelist

    >>> padd([59, 101, 0, 0, 156, 17, 1, 1, 3], 16)
    [59, 101, 0, 0, 156, 17, 1, 1, 3, 255, 255, 255, 255, 255, 255, 255]
    >>> padd([59, 101, 0, 0, 156, 17, 1, 1, 3], 12, '80')
    [59, 101, 0, 0, 156, 17, 1, 1, 3, 128, 128, 128]
    >>> padd([59, 101, 0, 0, 156, 17, 1, 1, 3], 8)
    [59, 101, 0, 0, 156, 17, 1, 1, 3]
    """

    return bytelist + [int(padding, 16)] * (length - len(bytelist))


def toASCIIBytes(stringtoconvert: str) -> list[int]:
    """Convert a string to a list of UTF-8 encoded bytes.

    @param stringtoconvert: the string to convert into a byte list

    @return: a byte list of the ASCII codes of the string characters

    L{toASCIIBytes()} is the reverse of L{toASCIIString()}

    >>> toASCIIBytes("Number 101")
    [78, 117, 109, 98, 101, 114, 32, 49, 48, 49]
    """

    return list(stringtoconvert.encode("utf-8"))


def toASCIIString(bytelist: list[int]) -> str:
    """Convert a list of integers in the range ``[32, 127]`` to a string.

    Integer values outside the range ``[32, 127]`` are replaced with a period.

    @param bytelist: list of ASCII bytes to convert into a string

    @return: a string from the ASCII code list

    L{toASCIIString()} is the reverse of L{toASCIIBytes()}

    >>> toASCIIString([0x4E,0x75,0x6D,0x62,0x65,0x72,0x20,0x31,0x30,0x31])
    'Number 101'
    >>> toASCIIString([0x01, 0x20, 0x80, 0x7E, 0xF0])
    ". .~."
    """

    return "".join(chr(c) if 32 <= c <= 127 else "." for c in bytelist)


def toBytes(bytestring: str) -> list[int]:
    """Convert a string of hexadecimal characters to a list of integers.

    @param bytestring: a byte string

    >>> toBytes("3B 65 00 00 9C 11 01 01 03")
    [59, 101, 0, 0, 156, 17, 1, 1, 3]
    >>> toBytes("3B6500009C11010103")
    [59, 101, 0, 0, 156, 17, 1, 1, 3]
    >>> toBytes("3B6500   009C1101  0103")
    [59, 101, 0, 0, 156, 17, 1, 1, 3]
    """

    try:
        return list(bytes.fromhex(bytestring))
    except ValueError as exc:
        raise TypeError("not a string representing a list of bytes") from exc


# GSM3.38 character conversion table.
__dic_GSM_3_38__ = {
    "@": 0x00,  # @ At symbol
    "£": 0x01,  # £ Britain pound symbol
    "$": 0x02,  # $ Dollar symbol
    "¥": 0x03,  # ¥ Yen symbol
    "è": 0x04,  # è e accent grave
    "é": 0x05,  # é e accent aigu
    "ù": 0x06,  # ù u accent grave
    "ì": 0x07,  # ì i accent grave
    "ò": 0x08,  # ò o accent grave
    "Ç": 0x09,  # Ç C majuscule cedille
    "\n": 0x0A,  # LF Line Feed
    "Ø": 0x0B,  # Ø O majuscule barré
    "ø": 0x0C,  # ø o minuscule barré
    "\r": 0x0D,  # CR Carriage Return
    "Å": 0x0E,  # Å Angstroem majuscule
    "å": 0x0F,  # å Angstroem minuscule
    "Δ": 0x10,  # Δ Greek letter delta
    "_": 0x11,  # underscore
    "Φ": 0x12,  # Φ Greek letter phi
    "Γ": 0x13,  # Γ Greek letter gamma
    "Λ": 0x14,  # Λ Greek letter lambda
    "Ω": 0x15,  # Ω Greek letter omega
    "Π": 0x16,  # Π Greek letter pi
    "Ψ": 0x17,  # Ψ Greek letter psi
    "Σ": 0x18,  # Σ Greek letter sigma
    "Θ": 0x19,  # Θ Greek letter theta
    "Ξ": 0x1A,  # Ξ Greek letter xi
    #    0x1B               maps to extension table
    "Æ": 0x1C,  # Æ majuscule ae
    "æ": 0x1D,  # æ minuscule ae
    "ß": 0x1E,  # ß s dur allemand
    "É": 0x1F,  # É majuscule é
    " ": 0x20,
    "!": 0x21,
    '"': 0x22,  # guillemet
    "#": 0x23,
    "¤": 0x24,  # ¤ carré
    #
    #    0x25 ... 0x3F    # % ... ?
    #
    "¡": 0x40,  # ¡ point d'exclamation renversé
    #
    #    0x41 ... 0x5A    # A ... Z
    #
    "Ä": 0x5B,  # Ä majuscule A trema
    "Ö": 0x5C,  # Ö majuscule O trema
    "Ñ": 0x5D,  # Ñ majuscule N tilde espagnol
    "Ü": 0x5E,  # Ü majuscule U trema
    "§": 0x5F,  # § signe paragraphe
    "¿": 0x60,  # ¿ point interrogation renversé
    #
    #    0x61 ... 0x7A    # a ... z
    #
    "ä": 0x7B,  # ä minuscule a trema
    "ö": 0x7C,  # ö minuscule o trema
    "ñ": 0x7D,  # ñ minuscule n tilde espagnol
    "ü": 0x7E,  # ü minuscule u trema
    "à": 0x7F,  # à a accent grave
}


def toGSM3_38Bytes(stringtoconvert: str | bytes) -> list[int]:
    """Returns a list of bytes from a string using GSM 3.38 conversion table.

    @param stringtoconvert:     string to convert

    @return: a list of bytes

    >>> toGSM3_38Bytes("@ùPascal")
    [0, 6, 80, 97, 115, 99, 97, 108]
    """
    if isinstance(stringtoconvert, bytes):
        stringtoconvert = stringtoconvert.decode("iso8859-1")

    result = []
    for char in stringtoconvert:
        if ("%" <= char <= "?") or ("A" <= char <= "Z") or ("a" <= char <= "z"):
            result.append(ord(char))
        else:
            result.append(__dic_GSM_3_38__[char])
    return result


def toHexString(data: list[int] | None = None, output_format: int = 0) -> str:
    """Convert a list of integers to a formatted string of hexadecimal.

    Integers larger than 255 will be truncated to two-byte hexadecimal pairs.

    @param data:   a list of bytes to stringify,
                e.g. [59, 22, 148, 32, 2, 1, 0, 0, 13]
    @param output_format: a logical OR of
      - COMMA: add a comma between bytes
      - HEX: add the 0x chars before bytes
      - UPPERCASE: use 0X before bytes (need HEX)
      - PACK: remove blanks

    >>> vals = [0x3B, 0x65, 0x00, 0x00, 0x9C, 0x11, 0x01, 0x01, 0x03]
    >>> toHexString(vals)
    '3B 65 00 00 9C 11 01 01 03'
    >>> toHexString(vals, COMMA)
    '3B, 65, 00, 00, 9C, 11, 01, 01, 03'
    >>> toHexString(vals, HEX)
    '0x3B 0x65 0x00 0x00 0x9C 0x11 0x01 0x01 0x03'
    >>> toHexString(vals, HEX | COMMA)
    '0x3B, 0x65, 0x00, 0x00, 0x9C, 0x11, 0x01, 0x01, 0x03'
    >>> toHexString(vals, PACK)
    '3B6500009C11010103'
    >>> toHexString(vals, HEX | UPPERCASE)
    '0X3B 0X65 0X00 0X00 0X9C 0X11 0X01 0X01 0X03'
    >>> toHexString(vals, HEX | UPPERCASE | COMMA)
    '0X3B, 0X65, 0X00, 0X00, 0X9C, 0X11, 0X01, 0X01, 0X03'
    """

    if not (data is None or isinstance(data, list)):
        raise TypeError("not a list of bytes")

    if not data:
        return ""

    pformat = "%-0.2X"
    separator = ""
    if COMMA & output_format:
        separator = ","
    if not PACK & output_format:
        separator += " "
    if HEX & output_format:
        if UPPERCASE & output_format:
            pformat = "0X" + pformat
        else:
            pformat = "0x" + pformat
    return separator.join(pformat % (a & 0xFF) for a in data).rstrip()


def HexListToBinString(hexlist: list[int]) -> str:
    """Deprecated. Use `bytes(hexlist).decode("utf-8")` or similar.

    >>> HexListToBinString([78, 117, 109, 98, 101, 114, 32, 49, 48, 49])
    'Number 101'
    """

    warnings.warn(
        'Use `bytes(hexlist).decode("utf-8")` or similar.',
        DeprecationWarning,
    )
    return bytes(hexlist).decode("utf-8")


def BinStringToHexList(binstring: str) -> list[int]:
    """Deprecated. Use `list(binstring.encode("utf-8"))` or similar.

    >>> BinStringToHexList("Number 101")
    [78, 117, 109, 98, 101, 114, 32, 49, 48, 49]
    """

    warnings.warn(
        'Use `list(binstring.encode("utf-8"))` or similar.',
        DeprecationWarning,
    )
    return list(binstring.encode("utf-8"))


hl2bs = HexListToBinString
bs2hl = BinStringToHexList



"""The error checking chain is a list of status word
(sw1, sw2) error check strategies.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from sys import exc_info


class ErrorCheckingChain:
    """The error checking chain is a list of response apdu status word
    (sw1, sw2) error check strategies. Each strategy in the chain is
    called until an error is detected. A L{smartcard.sw.SWExceptions}
    exception is raised when an error is detected. No exception is
    raised if no error is detected.

    Implementation derived from Bruce Eckel, Thinking in Python. The
    L{ErrorCheckingChain} implements the Chain Of Responsibility design
    pattern.
    """

    def __init__(self, chain, strategy):
        """constructor. Appends a strategy to the L{ErrorCheckingChain}
        chain."""
        self.strategy = strategy
        self.chain = chain
        self.chain.append(self)
        self.excludes = []

    def next(self):
        """Returns next error checking strategy."""
        # Where this link is in the chain:
        location = self.chain.index(self)
        if not self.end():
            return self.chain[location + 1]
        return None

    def addFilterException(self, exClass):
        """Add an exception filter to the error checking chain.

        @param exClass:    the exception to exclude, e.g.
        L{smartcard.sw.SWExceptions.WarningProcessingException} A filtered
        exception will not be raised when the sw1,sw2 conditions that
        would raise the exception are met.
        """

        self.excludes.append(exClass)
        if self.end():
            return
        self.next().addFilterException(exClass)

    def end(self):
        """Returns True if this is the end of the error checking
        strategy chain."""
        return self.chain.index(self) + 1 >= len(self.chain)

    def __call__(self, data, sw1, sw2):
        """Called to test data, sw1 and sw2 for error on the chain."""
        try:
            self.strategy(data, sw1, sw2)
        except tuple(self.excludes):
            # The following additional filter may look redundant, it isn't.
            # It checks that type(exc) is *equal* to any of self.excludes,
            # rather than equal-or-subclass to any of self.excludes.
            # This maintains backward compatibility with the behaviour of
            # pyscard <= 1.6.16.
            # if exception is filtered, return
            for exception in self.excludes:
                if exception == exc_info()[0]:
                    return None
            # otherwise reraise exception
            raise

        # if not done, call next strategy
        if self.end():
            return None
        return self.next()(data, sw1, sw2)



"""ISO7816-4 sw1 only error checker.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import smartcard.sw.SWExceptions
from smartcard.sw.ErrorChecker import ErrorChecker

# pylint: disable=too-few-public-methods

iso7816_4SW1 = {
    0x62: smartcard.sw.SWExceptions.WarningProcessingException,
    0x63: smartcard.sw.SWExceptions.WarningProcessingException,
    0x64: smartcard.sw.SWExceptions.ExecutionErrorException,
    0x65: smartcard.sw.SWExceptions.ExecutionErrorException,
    0x66: smartcard.sw.SWExceptions.SecurityRelatedException,
    0x67: smartcard.sw.SWExceptions.CheckingErrorException,
    0x68: smartcard.sw.SWExceptions.CheckingErrorException,
    0x69: smartcard.sw.SWExceptions.CheckingErrorException,
    0x6A: smartcard.sw.SWExceptions.CheckingErrorException,
    0x6B: smartcard.sw.SWExceptions.CheckingErrorException,
    0x6C: smartcard.sw.SWExceptions.CheckingErrorException,
    0x6D: smartcard.sw.SWExceptions.CheckingErrorException,
    0x6E: smartcard.sw.SWExceptions.CheckingErrorException,
    0x6F: smartcard.sw.SWExceptions.CheckingErrorException,
}


class ISO7816_4_SW1ErrorChecker(ErrorChecker):
    """ISO7816-4 error checker based on status word sw1 only.

    This error checker raises the following exceptions:
      - sw1 sw2
      - 62  any   L{WarningProcessingException}
      - 63  any   L{WarningProcessingException}
      - 64  any   L{ExecutionErrorException}
      - 65  any   L{ExecutionErrorException}
      - 66  any   L{SecurityRelatedException}
      - 67  any   L{CheckingErrorException}
      - 68  any   L{CheckingErrorException}
      - 69  any   L{CheckingErrorException}
      - 6a  any   L{CheckingErrorException}
      - 6b  any   L{CheckingErrorException}
      - 6c  any   L{CheckingErrorException}
      - 6d  any   L{CheckingErrorException}
      - 6e  any   L{CheckingErrorException}
      - 6f  any   L{CheckingErrorException}
    """

    def __call__(self, data, sw1, sw2):
        """Called to test data, sw1 and sw2 for error.

        @param data:       apdu response data
        @param sw1:        apdu data status word 1
        @param sw2:        apdu data status word 2
        """
        if sw1 in iso7816_4SW1:
            exception = iso7816_4SW1[sw1]
            raise exception(data, sw1, sw2)


if __name__ == "__main__":
    # Small sample illustrating the use of ISO7816_4_SW1ErrorChecker.
    ecs = ISO7816_4_SW1ErrorChecker()
    ecs([], 0x90, 0x00)
    try:
        ecs([], 0x66, 0x80)
    except smartcard.sw.SWExceptions.SecurityRelatedException as e:
        print(str(e) + f" {e.sw1:x} {e.sw2:x}")



"""Card class.

__author__ = "gemalto https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.reader.Reader import Reader
from smartcard.System import readers
from smartcard.util import toHexString


class Card:
    """Card class."""

    def __init__(self, reader, atr):
        """Card constructor.
        @param reader: reader in which the card is inserted
        @param atr: ATR of the card"""
        self.reader = reader
        self.atr = atr

    def __repr__(self):
        """Return a string representing the Card (atr and reader
        concatenation)."""
        return toHexString(self.atr) + " / " + str(self.reader)

    def __eq__(self, other):
        """Return True if self==other (same reader and same atr).
        Return False otherwise."""
        if isinstance(other, Card):
            return self.atr == other.atr and repr(self.reader) == repr(other.reader)

        return False

    def __ne__(self, other):
        """Return True if self!=other (same reader and same atr).Returns
        False otherwise."""
        return not self.__eq__(other)

    def __hash__(self):
        """Returns a hash value for this object (str(self) is unique)."""
        return hash(str(self))

    def createConnection(self):
        """Return a CardConnection to the Card object."""
        readerobj = None
        if isinstance(self.reader, Reader):
            readerobj = self.reader
        elif isinstance(self.reader, str):
            for reader in readers():
                if self.reader == str(reader):
                    readerobj = reader

        if readerobj:
            return readerobj.createConnection()

        # raise CardConnectionException(
        # 'not a valid reader: ' + str(self.reader))
        return None



"""
from Thinking in Python, Bruce Eckel
https://python-3-patterns-idioms-test.readthedocs.io/en/latest/Observer.html

(c) Copyright 2008, Creative Commons Attribution-Share Alike 3.0.

Class support for "observer" pattern.

The observer class is the base class
for all smartcard package observers.
"""

from __future__ import annotations

import typing

from smartcard.Synchronization import Synchronization, synchronize

# pylint: disable=too-few-public-methods


class Observer:
    """Observer"""

    def update(self, observable: Observable, handlers: typing.Any) -> None:
        """Called when the observed object is
        modified. You call an Observable object's
        notifyObservers method to notify all the
        object's observers of the change."""


class Observable(Synchronization):
    """Observable"""

    def __init__(self) -> None:
        super().__init__()
        self.obs: list[Observer] = []
        self.changed = 0

    def addObserver(self, observer: Observer) -> None:
        """Add an observer"""
        if observer not in self.obs:
            self.obs.append(observer)

    def deleteObserver(self, observer: Observer) -> None:
        """Remove an observer"""
        self.obs.remove(observer)

    def notifyObservers(self, handlers: typing.Any = None) -> None:
        """If 'changed' indicates that this object
        has changed, notify all its observers, then
        call clearChanged(). Each observer has its
        update() called with two arguments: this
        observable object and the generic 'handlers'."""

        with self.mutex:
            if not self.changed:
                return
            # Make a copy of the observer list.
            observers = self.obs.copy()
            self.changed = 0

        # Update observers
        for observer in observers:
            observer.update(self, handlers)

    def deleteObservers(self) -> None:
        """Remove all observers"""
        self.obs = []

    def setChanged(self) -> None:
        """Set the change flag"""
        self.changed = 1

    def clearChanged(self) -> None:
        """Clear the change flag"""
        self.changed = 0

    def hasChanged(self) -> int:
        """Somethig has changed?"""
        return self.changed

    def countObservers(self) -> int:
        """Return the number of Observers"""
        return len(self.obs)


synchronize(
    Observable,
    "addObserver deleteObserver deleteObservers "
    + "setChanged clearChanged hasChanged "
    + "countObservers",
)



"""
Smartcard Session.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.Exceptions import InvalidReaderException, NoReadersException
from smartcard.PassThruCardService import PassThruCardService
from smartcard.System import readers


class Session:
    """The Session object enables programmers to transmit APDU to smartcards.

    This is an example of use of the Session object:

    >>> import smartcard
    >>> reader=smartcard.listReaders()
    >>> s = smartcard.Session(reader[0])
    >>> SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
    >>> DF_TELECOM = [0x7F, 0x10]
    >>> data, sw1, sw2 = s.sendCommandAPDU(SELECT+DF_TELECOM)
    >>> print(data, sw1, sw2)
    >>> s.close()
    >>> print(`s`)
    """

    def __init__(self, readerName=None):
        """Session constructor. Initializes a smart card session and
        connect to the card.

        @param readerName: reader to connect to; default is first PCSC reader
        """

        # if reader name not given, select first reader
        if readerName is None:
            if len(readers()) > 0:
                self.reader = readers()[0]
                self.readerName = repr(self.reader)
            else:
                raise NoReadersException()

        # otherwise select reader from name
        else:
            self.readerName = readerName
            for reader in readers():
                if readerName == str(reader):
                    self.reader = reader
                    self.readerName = repr(self.reader)

        try:
            self.reader
        except AttributeError as exc:
            raise InvalidReaderException(self.readerName) from exc

        # open card connection and bind PassThruCardService
        cc = self.reader.createConnection()
        self.cs = PassThruCardService(cc)
        self.cs.connection.connect()

    def close(self):
        """Close the smartcard session.

        Closing a session will disconnect from the card."""
        self.cs.connection.disconnect()

    def sendCommandAPDU(self, command):
        """Send an APDU command to the connected smartcard.

        @param command: list of APDU bytes, e.g. [0xA0, 0xA4, 0x00, 0x00, 0x02]

        @return: a tuple (response, sw1, sw2) where
                response is the APDU response
                sw1, sw2 are the two status words
        """

        response, sw1, sw2 = self.cs.connection.transmit(command)

        if len(response) > 2:
            response.append(sw1)
            response.append(sw2)
        return response, sw1, sw2

    def getATR(self):
        """Returns the ATR of the connected card."""
        return self.cs.connection.getATR()

    def __repr__(self):
        """Returns a string representation of the session."""
        return f"<Session instance: readerName={self.readerName}>"


if __name__ == "__main__":
    # Small sample illustrating the use of Session.py.
    pass



"""ReaderFactory: creates smartcard readers.

__author__ = "gemalto https://www.gemalto.com/"

Factory pattern implementation borrowed from
Thinking in Python, Bruce Eckel,
http://mindview.net/Books/TIPython

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import importlib

from smartcard.pcsc.PCSCReader import PCSCReader


class ReaderFactory:
    """Class to create readers from reader type id."""

    factories = {}
    factorymethods = [PCSCReader.readers]

    # A Template Method:
    @staticmethod
    def createReader(clazz: str, readername: str):
        """Static method to create a reader from a reader clazz.

        @param clazz:      the reader class name
        @param readername: the reader name
        """

        if clazz not in ReaderFactory.factories:
            module_name, _, class_name = clazz.rpartition(".")
            imported_module = importlib.import_module(module_name)
            imported_class = getattr(imported_module, class_name)
            ReaderFactory.factories[clazz] = imported_class.Factory()

        return ReaderFactory.factories[clazz].create(readername)

    @staticmethod
    def readers(groups=None):
        """Return the list of readers"""
        if groups is None:
            groups = []
        zreaders = []
        for fm in ReaderFactory.factorymethods:
            zreaders += fm(groups)
        return zreaders



"""wxpython smartcard utility module.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import os.path
import sys


def main_is_frozen():
    return hasattr(sys, "frozen") or hasattr(sys, "importers")


ICO_SMARTCARD = None
ICO_READER = None

# running from a script, i.e. not running from standalone exe built with py2exe
if not main_is_frozen():
    ICO_SMARTCARD = os.path.join(
        os.path.dirname(__file__), "resources", "smartcard.ico"
    )
    ICO_READER = os.path.join(os.path.dirname(__file__), "resources", "reader.ico")

# running from a standalone exe built with py2exe
# resources expected images directory
else:
    if os.path.exists(os.path.join("images", "smartcard.ico")):
        ICO_SMARTCARD = os.path.join("images", "smartcard.ico")
    if os.path.exists(os.path.join("images", "reader.ico")):
        ICO_READER = os.path.join("images", "reader.ico")



"""Abstract CarType.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.Exceptions import InvalidATRMaskLengthException
from smartcard.System import readers
from smartcard.util import toHexString

# pylint: disable=too-few-public-methods


class CardType:
    """Abstract base class for CardTypes.

    Known subclasses: L{smartcard.CardType.AnyCardType}
    L{smartcard.CardType.ATRCardType}."""

    def __init__(self):
        """CardType constructor."""

    def matches(self, atr, reader=None):
        """Returns true if atr and card connected match the L{CardType}.

        @param atr:    the atr to check for matching
        @param reader: the reader (optional); default is None

        The reader can be used in some subclasses to do advanced
        matching that require connecting to the card."""


class AnyCardType(CardType):
    """The AnyCardType matches any card."""

    def matches(self, atr, reader=None):
        """Always returns true, i.e. AnyCardType matches any card.

        @param atr:    the atr to check for matching
        @param reader: the reader (optional); default is None"""
        return True


class ATRCardType(CardType):
    """The ATRCardType defines a card from an ATR and a mask."""

    def __init__(self, atr, mask=None):
        """ATRCardType constructor.
        @param atr:    the ATR of the CardType
        @param mask:   an optional mask to be applied to the ATR for
            L{CardType} matching default is None
        """
        super().__init__()
        self.atr = list(atr)
        self.mask = mask
        if mask is None:
            self.maskedatr = self.atr
        else:
            if len(self.atr) != len(self.mask):
                raise InvalidATRMaskLengthException(toHexString(mask))
            self.maskedatr = list(map(lambda x, y: x & y, self.atr, self.mask))

    def matches(self, atr, reader=None):
        """Returns true if the atr matches the masked CardType atr.

        @param atr:    the atr to check for matching
        @param reader: the reader (optional); default is None

        When atr is compared to the CardType ATR, matches returns true if
        and only if CardType.atr & CardType.mask = atr & CardType.mask,
        where & is the bitwise logical AND."""

        if len(atr) != len(self.atr):
            return not True

        if self.mask is not None:
            maskedatr = list(map(lambda x, y: x & y, list(atr), self.mask))
        else:
            maskedatr = atr
        return self.maskedatr == maskedatr


if __name__ == "__main__":
    # Small sample illustrating the use of CardType.py.
    r = readers()
    print(r)
    connection = r[0].createConnection()
    connection.connect()
    atrct = ATRCardType([0x3B, 0x16, 0x94, 0x20, 0x02, 0x01, 0x00, 0x00, 0x0D])
    print(atrct.matches(connection.getATR()))



"""PCSCReaderGroups organizes smartcard readers as groups.

__author__ = "gemalto https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.pcsc.PCSCExceptions import (
    EstablishContextException,
    ListReadersException,
    ReleaseContextException,
)
from smartcard.reader.ReaderGroups import innerreadergroups, readergroups
from smartcard.scard import (
    SCARD_S_SUCCESS,
    SCARD_SCOPE_USER,
    SCardEstablishContext,
    SCardForgetReaderGroup,
    SCardGetErrorMessage,
    SCardIntroduceReaderGroup,
    SCardListReaderGroups,
    SCardReleaseContext,
    error,
)

# pylint: disable=too-few-public-methods


class pcscinnerreadergroups(innerreadergroups):
    """Smartcard PCSC readers groups inner class.

    The PCSCReaderGroups singleton manages the creation of the unique
    instance of this class.
    """

    def __init__(self, initlist=None):
        """Constructor."""
        innerreadergroups.__init__(self, initlist)
        self.unremovablegroups = ["SCard$DefaultReaders"]

    def getreadergroups(self):
        """Returns the list of smartcard reader groups."""
        innerreadergroups.getreadergroups(self)

        hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if hresult != SCARD_S_SUCCESS:
            raise EstablishContextException(hresult)
        hresult, readers = SCardListReaderGroups(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise ListReadersException(hresult)
        hresult = SCardReleaseContext(hcontext)
        if hresult != SCARD_S_SUCCESS:
            raise ReleaseContextException(hresult)
        return readers

    def addreadergroup(self, newgroup):
        """Add a reader group"""

        hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if SCARD_S_SUCCESS != hresult:
            raise error("Failed to establish context: " + SCardGetErrorMessage(hresult))
        try:
            hresult = SCardIntroduceReaderGroup(hcontext, newgroup)
            if SCARD_S_SUCCESS != hresult:
                raise error(
                    "Unable to introduce reader group: " + SCardGetErrorMessage(hresult)
                )

            innerreadergroups.addreadergroup(self, newgroup)

        finally:
            hresult = SCardReleaseContext(hcontext)
            if SCARD_S_SUCCESS != hresult:
                raise error(
                    "Failed to release context: " + SCardGetErrorMessage(hresult)
                )

    def removereadergroup(self, group):
        """Remove a reader group"""

        hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if SCARD_S_SUCCESS != hresult:
            raise error("Failed to establish context: " + SCardGetErrorMessage(hresult))
        try:
            hresult = SCardForgetReaderGroup(hcontext, group)
            if hresult != SCARD_S_SUCCESS:
                raise error(
                    "Unable to forget reader group: " + SCardGetErrorMessage(hresult)
                )

            innerreadergroups.removereadergroup(self, group)

        finally:
            hresult = SCardReleaseContext(hcontext)
            if SCARD_S_SUCCESS != hresult:
                raise error(
                    "Failed to release context: " + SCardGetErrorMessage(hresult)
                )


class PCSCReaderGroups(readergroups):
    """PCSC readers groups."""

    def __init__(self, initlist=None):
        """Create a single instance of pcscinnerreadergroups on first call"""
        self.innerclazz = pcscinnerreadergroups
        readergroups.__init__(self, initlist)


if __name__ == "__main__":
    print(PCSCReaderGroups().getreadergroups())



"""Simple wxpython frame for smart card application.

__author__ = "gemalto https://www.gemalto.com/"
__date__ = "November 2006"
__version__ = "1.4.0"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import os.path

import wx

import smartcard
import smartcard.wx
from smartcard.wx import APDUTracerPanel, CardAndReaderTreePanel, ReaderToolbar
from smartcard.wx.SimpleSCardAppEventObserver import SimpleSCardAppEventObserver

[
    wxID_SIMPLESCARDAPP_FRAME,
] = [wx.NewId() for x in range(1)]


class BlankPanel(wx.Panel, SimpleSCardAppEventObserver):
    """A blank panel in case no panel is provided to SimpleSCardApp."""

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, -1)
        sizer = wx.GridSizer(1, 1, 0)
        self.SetSizer(sizer)
        self.SetAutoLayout(True)


class TreeAndUserPanelPanel(wx.Panel):
    """The panel that contains the Card/Reader TreeCtrl
    and the user provided Panel."""

    def __init__(self, parent, apppanelclass, appstyle):
        """
        Constructor. Creates the panel with two panels:
          - the left-hand panel is holding the smartcard and/or reader tree
          - the right-hand panel is holding the application dialog

        @param apppanelclass: the class of the panel to instantiate in the
                    L{SimpleSCardAppFrame}
        @param appstyle: a combination of the following styles (bitwise or |)
           - TR_SMARTCARD: display a smartcard tree panel
           - TR_READER: display a reader tree panel
           - TB_SMARTCARD: display a smartcard toolbar
           - TB_SMARTCARD: display a reader toolbar
           - default is TR_DEFAULT = TR_SMARTCARD
        """
        wx.Panel.__init__(self, parent, -1)

        self.parent = parent
        self.selectedcard = None

        boxsizer = wx.BoxSizer(wx.HORIZONTAL)

        # create user dialog
        if None != apppanelclass:
            self.dialogpanel = apppanelclass(self)
        else:
            self.dialogpanel = BlankPanel(self)

        # create card/reader tree control
        if (
            appstyle & smartcard.wx.SimpleSCardApp.TR_SMARTCARD
            or appstyle & smartcard.wx.SimpleSCardApp.TR_READER
        ):
            self.readertreepanel = CardAndReaderTreePanel.CardAndReaderTreePanel(
                self, appstyle, self.dialogpanel
            )
            boxsizer.Add(self.readertreepanel, 1, wx.EXPAND | wx.ALL, 5)

        boxsizer.Add(self.dialogpanel, 2, wx.EXPAND | wx.ALL)

        if appstyle & smartcard.wx.SimpleSCardApp.TR_READER:
            self.Bind(
                wx.EVT_TREE_ITEM_ACTIVATED,
                self.OnActivateReader,
                self.readertreepanel.readertreectrl,
            )
            self.Bind(
                wx.EVT_TREE_SEL_CHANGED,
                self.OnSelectReader,
                self.readertreepanel.readertreectrl,
            )
            self.Bind(
                wx.EVT_TREE_ITEM_RIGHT_CLICK,
                self.OnReaderRightClick,
                self.readertreepanel.readertreectrl,
            )
            self.Bind(
                wx.EVT_TREE_ITEM_COLLAPSED,
                self.OnItemCollapsed,
                self.readertreepanel.readertreectrl,
            )

        if appstyle & smartcard.wx.SimpleSCardApp.TR_SMARTCARD:
            self.Bind(
                wx.EVT_TREE_ITEM_ACTIVATED,
                self.OnActivateCard,
                self.readertreepanel.cardtreectrl,
            )
            self.Bind(
                wx.EVT_TREE_SEL_CHANGED,
                self.OnSelectCard,
                self.readertreepanel.cardtreectrl,
            )
            self.Bind(
                wx.EVT_TREE_ITEM_RIGHT_CLICK,
                self.OnCardRightClick,
                self.readertreepanel.cardtreectrl,
            )

        self.SetSizer(boxsizer)
        self.SetAutoLayout(True)

    def ActivateCard(self, card):
        """Activate a card."""
        if not hasattr(card, "connection"):
            card.connection = card.createConnection()
            if None != self.parent.apdutracerpanel:
                card.connection.addObserver(self.parent.apdutracerpanel)
            card.connection.connect()
        self.dialogpanel.OnActivateCard(card)

    def DeactivateCard(self, card):
        """Deactivate a card."""
        if hasattr(card, "connection"):
            card.connection.disconnect()
            if None != self.parent.apdutracerpanel:
                card.connection.deleteObserver(self.parent.apdutracerpanel)
            delattr(card, "connection")
        self.dialogpanel.OnDeactivateCard(card)

    def OnActivateCard(self, event):
        """Called when the user activates a card in the tree."""
        item = event.GetItem()
        if item:
            itemdata = self.readertreepanel.cardtreectrl.GetItemData(item)
            if isinstance(itemdata, smartcard.Card.Card):
                self.ActivateCard(itemdata)
            else:
                self.dialogpanel.OnDeselectCard(itemdata)

    def OnActivateReader(self, event):
        """Called when the user activates a reader in the tree."""
        item = event.GetItem()
        if item:
            itemdata = self.readertreepanel.readertreectrl.GetItemData(item)
            if isinstance(itemdata, smartcard.Card.Card):
                self.ActivateCard(itemdata)
            elif isinstance(itemdata, smartcard.reader.Reader.Reader):
                self.dialogpanel.OnActivateReader(itemdata)
        event.Skip()

    def OnItemCollapsed(self, event):
        item = event.GetItem()
        self.readertreepanel.readertreectrl.Expand(item)

    def OnCardRightClick(self, event):
        """Called when user right-clicks a node in the card tree control."""
        item = event.GetItem()
        if item:
            itemdata = self.readertreepanel.cardtreectrl.GetItemData(item)
            if isinstance(itemdata, smartcard.Card.Card):
                self.selectedcard = itemdata
                if not hasattr(self, "connectID"):
                    self.connectID = wx.NewId()
                    self.disconnectID = wx.NewId()

                    self.Bind(wx.EVT_MENU, self.OnConnect, id=self.connectID)
                    self.Bind(wx.EVT_MENU, self.OnDisconnect, id=self.disconnectID)

                menu = wx.Menu()
                if not hasattr(self.selectedcard, "connection"):
                    menu.Append(self.connectID, "Connect")
                else:
                    menu.Append(self.disconnectID, "Disconnect")
                self.PopupMenu(menu)
                menu.Destroy()

    def OnReaderRightClick(self, event):
        """Called when user right-clicks a node in the reader tree control."""
        item = event.GetItem()
        if item:
            itemdata = self.readertreepanel.readertreectrl.GetItemData(item)
            if isinstance(itemdata, smartcard.Card.Card):
                self.selectedcard = itemdata
                if not hasattr(self, "connectID"):
                    self.connectID = wx.NewId()
                    self.disconnectID = wx.NewId()

                    self.Bind(wx.EVT_MENU, self.OnConnect, id=self.connectID)
                    self.Bind(wx.EVT_MENU, self.OnDisconnect, id=self.disconnectID)

                menu = wx.Menu()
                if not hasattr(self.selectedcard, "connection"):
                    menu.Append(self.connectID, "Connect")
                else:
                    menu.Append(self.disconnectID, "Disconnect")
                self.PopupMenu(menu)
                menu.Destroy()

    def OnConnect(self, event):
        if isinstance(self.selectedcard, smartcard.Card.Card):
            self.ActivateCard(self.selectedcard)

    def OnDisconnect(self, event):
        if isinstance(self.selectedcard, smartcard.Card.Card):
            self.DeactivateCard(self.selectedcard)

    def OnSelectCard(self, event):
        """Called when the user selects a card in the tree."""
        item = event.GetItem()
        if item:
            itemdata = self.readertreepanel.cardtreectrl.GetItemData(item)
            if isinstance(itemdata, smartcard.Card.Card):
                self.dialogpanel.OnSelectCard(itemdata)
            else:
                self.dialogpanel.OnDeselectCard(itemdata)

    def OnSelectReader(self, event):
        """Called when the user selects a reader in the tree."""
        item = event.GetItem()
        if item:
            itemdata = self.readertreepanel.readertreectrl.GetItemData(item)
            if isinstance(itemdata, smartcard.Card.Card):
                self.dialogpanel.OnSelectCard(itemdata)
            elif isinstance(itemdata, smartcard.reader.Reader.Reader):
                self.dialogpanel.OnSelectReader(itemdata)
            else:
                self.dialogpanel.OnDeselectCard(itemdata)


class SimpleSCardAppFrame(wx.Frame):
    """The main frame of the simple smartcard application."""

    def __init__(
        self,
        appname,
        apppanelclass,
        appstyle,
        appicon,
        pos=(-1, -1),
        size=(-1, -1),
    ):
        """
        Constructor. Creates the frame with two panels:
          - the left-hand panel is holding the smartcard and/or reader tree
          - the right-hand panel is holding the application dialog

        @param appname: name of the application
        @param apppanelclass: the class of the panel to instantiate in the
                       L{SimpleSCardAppFrame}
        @param appstyle: a combination of the following styles (bitwise or |)
          - TR_SMARTCARD: display a smartcard tree panel
          - TR_READER: display a reader tree panel
          - TB_SMARTCARD: display a smartcard toolbar
          - TB_SMARTCARD: display a reader toolbar
          - PANEL_APDUTRACER: display an APDU tracer panel
          - default is TR_DEFAULT = TR_SMARTCARD
        @param pos: the application position as a (x,y) tuple; default is (-1,-1)
        @param size: the application window size as a (x,y) tuple; default is (-1,-1)
        """
        wx.Frame.__init__(
            self,
            None,
            wxID_SIMPLESCARDAPP_FRAME,
            appname,
            pos=pos,
            size=size,
            style=wx.DEFAULT_FRAME_STYLE,
        )

        if appicon:
            _icon = wx.Icon(appicon, wx.BITMAP_TYPE_ICO)
            self.SetIcon(_icon)
        elif os.path.exists(smartcard.wx.ICO_SMARTCARD):
            _icon = wx.Icon(smartcard.wx.ICO_SMARTCARD, wx.BITMAP_TYPE_ICO)
            self.SetIcon(_icon)

        boxsizer = wx.BoxSizer(wx.VERTICAL)
        self.treeuserpanel = TreeAndUserPanelPanel(self, apppanelclass, appstyle)
        boxsizer.Add(self.treeuserpanel, 3, wx.EXPAND | wx.ALL)

        # create a toolbar if required
        if appstyle & smartcard.wx.SimpleSCardApp.TB_SMARTCARD:
            self.toolbar = ReaderToolbar.ReaderToolbar(self)
            self.SetToolBar(self.toolbar)
        else:
            self.toolbar = None

        # create an apdu tracer console if required
        if appstyle & smartcard.wx.SimpleSCardApp.PANEL_APDUTRACER:
            self.apdutracerpanel = APDUTracerPanel.APDUTracerPanel(self)
            boxsizer.Add(self.apdutracerpanel, 1, wx.EXPAND | wx.ALL)
        else:
            self.apdutracerpanel = None

        self.SetSizer(boxsizer)
        self.SetAutoLayout(True)

        self.Bind(wx.EVT_CLOSE, self.OnCloseFrame)
        if appstyle & smartcard.wx.SimpleSCardApp.TB_SMARTCARD:
            self.Bind(
                wx.EVT_COMBOBOX, self.OnReaderComboBox, self.toolbar.readercombobox
            )

    def OnCloseFrame(self, evt):
        """Called when frame is closed, i.e. on wx.EVT_CLOSE"""
        evt.Skip()

    def OnExit(self, evt):
        """Called when frame application exits."""
        self.Close(True)
        evt.Skip()

    def OnReaderComboBox(self, event):
        """Called when the user activates a reader in the toolbar combo box."""
        cb = event.GetEventObject()
        reader = cb.GetClientData(cb.GetSelection())
        if isinstance(reader, smartcard.reader.Reader.Reader):
            self.treeuserpanel.dialogpanel.OnActivateReader(reader)
